// Copyright (c) 2015, Giliam de Carpentier
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the 
// documentation and/or other materials provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED 
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Based on code from the Ogre3D SDK, Copyright (c) 2000-2012 Torus Knot Software Ltd.

#pragma once

#include "Ogre.h"
#include <limits>

class CustomCameraMan
{
public:
	CustomCameraMan(Ogre::Camera* cam)
	: mCamera(0)
	, mTopSpeed(250)
	, mVelocity(Ogre::Vector3::ZERO)
	, mOrbiting(false)
	, mLockPitch(false)
	, mLockYaw(false)
	, mGoingForward(false)
	, mGoingBack(false)
	, mGoingLeft(false)
	, mGoingRight(false)
	, mGoingUp(false)
	, mGoingDown(false)
	, mFastMove(false)
	{
		setCamera(cam);
	}

	virtual ~CustomCameraMan() {}

	virtual void setCamera(Ogre::Camera* cam)
	{
		mCamera = cam;
		mPivotPosition = mCamera->getDerivedPosition();
	}

	virtual bool IsOrbiting()
	{
		return mOrbiting;
	}

	virtual Ogre::Camera* getCamera()
	{
		return mCamera;
	}

	virtual Ogre::Vector3 getPivotPosition()
	{
		return mPivotPosition;
	}

	virtual void setPivotPosition(const Ogre::Vector3& position)
	{
		mPivotPosition = position;
	}

	virtual void setLockPitch(bool lockPitch)
	{
		mLockPitch = lockPitch;
	}

	virtual void setLockYaw(bool lockYaw)
	{
		mLockYaw = lockYaw;
	}

	virtual void manualStop()
	{
		mGoingForward = false;
		mGoingBack = false;
		mGoingLeft = false;
		mGoingRight = false;
		mGoingUp = false;
		mGoingDown = false;
		mVelocity = Ogre::Vector3::ZERO;
	}

	virtual bool frameRenderingQueued(const Ogre::FrameEvent& evt)
	{
		// build our acceleration vector based on keyboard input composite
		Ogre::Vector3 accel = Ogre::Vector3::ZERO;
		if (mGoingForward) accel += mCamera->getDirection();
		if (mGoingBack) accel -= mCamera->getDirection();
		if (mGoingRight) accel += mCamera->getRight();
		if (mGoingLeft) accel -= mCamera->getRight();
		if (mGoingUp) accel += mCamera->getUp();
		if (mGoingDown) accel -= mCamera->getUp();

		// if accelerating, try to reach top speed in a certain time
		Ogre::Real topSpeed = mFastMove ? mTopSpeed * 20 : mTopSpeed;
		if (accel.squaredLength() != 0)
		{
			accel.normalise();
			mVelocity += accel * topSpeed * evt.timeSinceLastFrame * 10;
		}
		// if not accelerating, try to stop in a certain time
		else mVelocity -= mVelocity * evt.timeSinceLastFrame * 10;

		Ogre::Real tooSmall = std::numeric_limits<Ogre::Real>::epsilon();

		// keep camera velocity below top speed and above epsilon
		if (mVelocity.squaredLength() > topSpeed * topSpeed)
		{
			mVelocity.normalise();
			mVelocity *= topSpeed;
		}
		else if (mVelocity.squaredLength() < tooSmall * tooSmall)
			mVelocity = Ogre::Vector3::ZERO;

		mPivotPosition += mVelocity * evt.timeSinceLastFrame;

		float weight = 1.0f - Ogre::Math::Exp(-evt.timeSinceLastFrame * 22.0f);
		mCamera->setPosition(mPivotPosition);

		return true;
	}

	virtual void injectKeyDown(const OIS::KeyEvent& evt)
	{
		if (evt.key == OIS::KC_W || evt.key == OIS::KC_UP) mGoingForward = true;
		else if (evt.key == OIS::KC_S || evt.key == OIS::KC_DOWN) mGoingBack = true;
		else if (evt.key == OIS::KC_A || evt.key == OIS::KC_LEFT) mGoingLeft = true;
		else if (evt.key == OIS::KC_D || evt.key == OIS::KC_RIGHT) mGoingRight = true;
		else if (evt.key == OIS::KC_Q || evt.key == OIS::KC_PGUP) mGoingUp = true;
		else if (evt.key == OIS::KC_E || evt.key == OIS::KC_PGDOWN) mGoingDown = true;
		else if (evt.key == OIS::KC_LSHIFT) mFastMove = true;
	}

	virtual void injectKeyUp(const OIS::KeyEvent& evt)
	{
		if (evt.key == OIS::KC_W || evt.key == OIS::KC_UP) mGoingForward = false;
		else if (evt.key == OIS::KC_S || evt.key == OIS::KC_DOWN) mGoingBack = false;
		else if (evt.key == OIS::KC_A || evt.key == OIS::KC_LEFT) mGoingLeft = false;
		else if (evt.key == OIS::KC_D || evt.key == OIS::KC_RIGHT) mGoingRight = false;
		else if (evt.key == OIS::KC_Q || evt.key == OIS::KC_PGUP) mGoingUp = false;
		else if (evt.key == OIS::KC_E || evt.key == OIS::KC_PGDOWN) mGoingDown = false;
		else if (evt.key == OIS::KC_LSHIFT) mFastMove = false;
	}

	virtual void injectMouseMove(const OIS::MouseEvent& evt)
	{
		if (mOrbiting)   // yaw around the target, and pitch locally
		{
			if (!mLockYaw)
			{
				mCamera->yaw(Ogre::Degree(-evt.state.X.rel * 0.25f));
			}
			if (!mLockPitch)
			{
				mCamera->pitch(Ogre::Degree(-evt.state.Y.rel * 0.25f));
			}
		}
	}

	virtual void injectMouseDown(const OIS::MouseEvent& evt, OIS::MouseButtonID id)
	{
		if (id == OIS::MB_Right) mOrbiting = !mOrbiting;
	}

	virtual void injectMouseUp(const OIS::MouseEvent& evt, OIS::MouseButtonID id)
	{
		//if (id == OIS::MB_Left) mOrbiting = false;
	}

protected:
	Ogre::Camera* mCamera;
	Ogre::Vector3 mPivotPosition;
	Ogre::Real mTopSpeed;
	Ogre::Vector3 mVelocity;
	bool mOrbiting;
	bool mLockPitch;
	bool mLockYaw;
	bool mGoingForward;
	bool mGoingBack;
	bool mGoingLeft;
	bool mGoingRight;
	bool mGoingUp;
	bool mGoingDown;
	bool mFastMove;
};
